# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


FILTER_DICT = {
    "ASTOCKS": {"SH": "^SH6", "SZ": "^SZ00|^SZ30"},  # A股
    "BSTOCKS": {"SH": "^SH9", "SZ": "^SZ2"},  # B股
    "INDEX": {"SH": "^SH00", "SZ": "^SZ399"},  # 指数
    "ETF": {"SH": "^SH51", "SZ": "^SZ15|^SZ16"},
    "CB": {"SH": "^SH100|^SH110|^SSH112|^SH113", "SZ": "^SZ12"},  # 可转债
    "ALL": {"SH": "^SH", "SZ": "^SZ"},
}


def filter_by_symbol(df, symbol_type="ALL", market="ALL"):
    """根据 symbol 过滤返回符合条件的数据

    df 至少应包含 symbol 列，该列类型为字符串，内容规则如下：

        * SHxxxxxx    上海股票交易所代码
        * SZxxxxxx    深圳股票交易所代码

    symbol_type: ASTOCKS/BSTOCKS/INDEX/ETF/CB/ALL

        * ASTOCKS: A股股票
        * BSTOCKS: B股股票
        * INDEX: 指数
        * ETF: ETF基金
        * CB: CONVERTIBLE BONDS，可转债

    market: SZ/SH/ALL

        * SZ: 深圳股票交易所
        * SH: 上海股票交易所
        * ALL: 所有


    股票代码

    * 上海证券交易所

        ========   =========
        首位代码    产品定义
        ========   =========
        0           国债／指数
        00          上证指数、沪深300指数、中证指数
        １          债券
        ２          回购
        ３          期货
        ４          备用
        ５          基金／权证
        ６          A股
        ７          非交易业务（发行、权益分配）
        ８          备用
        ９          B股
        ========   =========

    * 深圳证券交易所

        ========   =========
        首位代码    产品定义
        ========   =========
        00          A股证券
        002~004     中小板
        1           债券
        2           B股
        30          创业板证券
        39          综合指数、成份指数
        ========   =========
    """
    if symbol_type == "ALL" and market == "ALL":
        return df
    if symbol_type not in FILTER_DICT:
        symbol_type = "ALL"
    if market in FILTER_DICT[symbol_type]:
        pat = FILTER_DICT[symbol_type][market]
    else:
        pat = "|".join(FILTER_DICT[symbol_type].values())
    return df[df.symbol.str.contains(pat)]



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict

import aiohttp

_http_session = None
_init_in_session = defaultdict()


def get_http_session(force=False):
    global _http_session
    if force or not _http_session:
        _init_in_session.clear()
        _http_session = aiohttp.ClientSession(
            headers={
                "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:62.0) Gecko/20100101 Firefox/62.0"
            }
        )
    return _http_session



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""提供选股器接口

信息获取通道包括：

* https://xueqiu.com/hq/screener/CN
* https://xueqiu.com/hq/screener/US

沪深选股参数说明

    对于不需要过滤但需要返回的字段请传递 `field=ALL` ，否则对应字段的值也不会返回

    * 字段一览，参数名为 field

      `参考字段一览列表 <../_static/picker_xq_fields.json>`_

      获取地址 https://xueqiu.com/stock/screener/fields.json?category=SH

      如下字段的参数名为 `filed.period` ，其中 period 为财务报表的结束日期::

          财务报表里的所有指标
          财务比率里的所有指标
          基本指标 - 净资产收益率(%)     roediluted.20180630
          基本指标 - 净利润(万)          netprofit.20180630
          基本指标 - 每股净资产          bps.20180630
          基本指标 - 每股收益            eps.20180630

    * 地域一览，选股参数字段名为 `areacode` ，值为 `keycode`

      `参考地域一览列表 <../_static/picker_xq_areas.json>`_

      获取地址 https://xueqiu.com/stock/screener/areas.json

    * 行业一览，选股参数字段名为 `indcode` ，值为 `level2code`

      `参考行业一览列表 <../_static/picker_xq_industries.json>`_

      获取地址 https://xueqiu.com/stock/screener/industries.json?category=SH

    * 交易所一览，选股参数字段名为 `exchange`

      SH/SZ  -- 沪市/深市

加载已保存的数据::

    import pandas as pd
    from pathlib import Path

    df = pd.read_json(Path('20181018165415.json')
    df.set_index(["symbol"], inplace=True)
    # df.drop_duplicates(inplace=True)
    df.apply(pd.to_numeric, errors="coerce") # convert to numertic types

    # df['day'] = date.today()
    # df["day"] = datetime.now().replace(hour=16, minute=0, second=0, microsecond=0)
    # set index
    # df.set_index(["symbol", "day"], inplace=True)
"""
import os
import time
import json
import copy
import asyncio
import logging
from pathlib import Path
from datetime import datetime

import click
import pandas as pd

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..env import _init_in_session
from ..utils import fetch_http_data, add_doc


logger = logging.getLogger(__file__)
__all__ = ["async_pick_stocks", "async_get_field_values", "pick_stocks"]


async def _init(session, force=False):
    if force or not _init_in_session.get("xueqiu"):
        _init_in_session["xueqiu"] = True
        await session.get("https://xueqiu.com")
    return True


async def async_get_field_values(session, field, return_df=True):
    """
    https://xueqiu.com/stock/screener/values.json?category=SH&field=pettm&_=1540617201640
    获取该字段值的列表
    """
    await _init(session)

    url = "https://xueqiu.com/stock/screener/values.json"
    params = {"category": "SH", "field": field}
    params["_"] = int(time.time() * 1000)
    data = []
    async with session.get(url, params=params) as resp:
        if resp.status != 200:
            logger.error("pick stock from %s failed: %s", url, resp.status)
            return None

        resp_json = await resp.json()
        if "list" not in resp_json:
            logger.error("pick stocks unknown result: %s", resp_json)
        else:
            data = resp_json["list"]

    if not return_df:
        return data

    return pd.DataFrame(data, columns=(field,))


async def async_pick_stocks(session, filter_dict=None, data_path=None, return_df=True):
    """
    雪球选股器接口，此接口支持的参数较多，具体请参考 API 文档

    返回一个 pd.DataFrame
    出错，返回 None

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param filter_dict: 需要的字段及过滤条件，支持的字段名见 `API 文档`
    :param data_path: 数据保存路径
    :param return_df: 是否返回 `pandas.DataFrame` 对象，False 返回原始数据
    :returns: 行情原始数据或带有行情数据的 `pandas.DataFrame` 对象，见 return_df 参数

    取值方法
    GET https://xueqiu.com/stock/screener/screen.json

    filter_dict:

        * category: SH/US  -- 沪深/美股   required
        * orderby:symbol   -- 排序字段    required
        * oder:desc        -- 升降排序    required
        * exchange: SH/SZ  -- 沪市/深市
        * areacode: 地域板块
        * indcode: 行业
        * filter_fileds: min_max
        * current:ALL     当前价
        * pct:ALL         本日涨跌幅
        * volume:ALL      本日成交量
        * page:1
        * size: 30
        * _:timestamp
    """
    await _init(session)

    page_size = 100000
    curr_page = page_cnt = 1
    params = {
        "category": "SH",
        "orderby": "symbol",
        "order": "desc",
        "current": "ALL",  # 当前价
        "pct": "ALL",  # 本日涨跌幅
        "volume": "ALL",  # 本日成交量
        "page": curr_page,
        "size": page_size,
        # "exchange": SH/SZ  -- 沪市/深市
        # "areacode": 地域板块
        # "indcode": 行业
        # "filter_fileds": min_max
    }
    if filter_dict:
        params.update(filter_dict)

    url = "https://xueqiu.com/stock/screener/screen.json"

    stock_list = []
    logger.info("start pick stocks, params: %s", params)
    while curr_page <= page_cnt:
        params["page"] = curr_page
        params["_"] = int(time.time() * 1000)

        logger.info("Fetching first page")
        async with session.get(url, params=params) as resp:
            if resp.status != 200:
                logger.error("pick stock from %s failed: %s", url, resp.status)
                return None

            resp_json = await resp.json()
            total_cnt = resp_json["count"]
            if "list" not in resp_json:
                logger.error("pick stocks unknown result: %s", resp_json)
                break

            page_cnt = total_cnt // page_size + 1 if total_cnt % page_size != 0 else 0
            if page_cnt != 1:
                stock_list.extend(resp_json["list"])
            else:
                stock_list = resp_json["list"]
        curr_page += 1

    if not stock_list:
        logger.warn("pick stock no result")
    logger.info("pick stocks from %s finish", url)

    if data_path:
        data_path = Path(data_path) / MODULE_DATA_DIR / "picker_xueqiu"
        os.makedirs(data_path, exist_ok=True)
        data_fname = "picker" + "-" + datetime.now().strftime("%Y%m%d%H%M%S") + ".json"
        data_file = data_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(stock_list, dataf, indent=4, ensure_ascii=False)

    if not return_df:
        return stock_list

    df = pd.DataFrame(stock_list)
    if stock_list:
        df.set_index(["symbol"], inplace=True)
        df.dropna(axis="columns", how="all", inplace=True)
        # convert to numertic types
        # df.apply(pd.to_numeric, errors="ignore")
    return df


@add_doc(async_pick_stocks.__doc__)
def pick_stocks(*args, **kwargs):
    ret = fetch_http_data(async_pick_stocks, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.option("-p/-np", "--print/--no-print", "show", default=True)
@click.argument("fields", nargs=-1)
@stock_cli_group.command("picker-xq")
@click.pass_context
def pick_stocks_cli(ctx, save_path, show, fields):
    """雪球选股器

    fields 选股筛选条件，具体支持的字段见API文档描述，
    格式为 field=val, 多个字段空格分割: areacode=CN110000 exchange=SH ...
    """
    if not save_path:
        save_path = ctx.obj["data_path"]

    kwargs = {}
    for field in fields:
        try:
            key, val = field.split("=")
        except ValueError:
            click.echo(f"{field} invalid skipped")
            continue
        kwargs[key] = val
    data = pick_stocks(kwargs, save_path)
    if show:
        click.echo(data)


if __name__ == "__main__":
    pick_stocks_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""本模块负责获取上市公司公告文件

信息获取通道：

    post http://www.cninfo.com.cn/cninfo-new/announcement/query

    post form::

        category: category_ndbg_szsh;category_bndbg_szsh;
        column:
        limit:
        pageNum: 1
        pageSize: 30
        searchkey: // 搜索关键字
        seDate: // 查询日期: 2018-07-28 或者 2018-03-13+~+2018-07-28
        sortName:
        sortType:
        stock: 300121
        tabName: fulltext

    category 类型

        * name="category_ndbg_szsh" title="年度报告"
        * name="category_bndbg_szsh" title="半年度报告"
        * name="category_yjdbg_szsh" title="一季度报告"
        * name="category_sjdbg_szsh" title="三季度报告"
        * name="category_scgkfx_szsh" title="首次公开发行及上市"
        * name="category_pg_szsh" title="配股"
        * name="category_zf_szsh" title="增发"
        * name="category_kzhz_szsh" title="可转换债券"
        * name="category_qzxg_szsh" title="权证相关公告"
        * name="category_qtrz_szsh" title="其他融资"
        * name="category_qyfpxzcs_szsh" title="权益及限制出售股份"
        * name="category_gqbd_szsh" title="股权变动"
        * name="category_jy_szsh" title="交易"
        * name="category_gddh_szsh" title="股东大会"
        * name="category_cqfxyj_szsh" title="澄清风险业绩预告"
        * name="category_tbclts_szsh" title="特别处理和退市"
        * name="category_bcgz_szsh" title="补充及更正"
        * name="category_zjjg_szsh" title="中介机构报告"
        * name="category_ssgszd_szsh" title="上市公司制度"
        * name="category_zqgg_szsh" title="债券公告"
        * name="category_qtzdsx_szsh" title="其它重大事项"
        * name="category_tzzgx_szsh" title="投资者关系信息"
        * name="category_dshgg_szsh" title="董事会公告"
        * name="category_jshgg_szsh" title="监事会公告"

加载已保存的数据::

    import json

    from pathlib import Path
    with Path("xxx.json").open(encoding="utf-8") as f:
        data = json.load(f)
"""
import os
import json
import asyncio
import logging
from pathlib import Path
from datetime import datetime, date
from dateutil.relativedelta import relativedelta

import click
import aiohttp

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..utils import parse_dt, fetch_http_data, add_doc


logger = logging.getLogger(__file__)
__all__ = ["get_announcements", "async_get_announcements"]


async def _init(session):
    return True


async def _get_one_announcement(session, url, fname):
    try:
        resp = await session.get(url)
        if resp.status != 200:
            logger.warning(
                "Download announcement %s from url failed: http %s", url, resp.status
            )
            return None
        raw_data = await resp.read()
    except (aiohttp.ServerTimeoutError, asyncio.TimerHandle):
        logger.warning("Download announcement %s failed：timeout", fname)
        return None
    with Path(fname).open("wb") as statf:
        statf.write(raw_data)
    return True


async def async_get_announcements(
    session,
    symbol,
    categories,
    data_path,
    start_date=None,
    end_date=None,
    search_key="",
):
    """获取公告文件

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param categories: 公告文件类别
    :param data_path: 数据保存路径
    :param start_date: 公共查询起始时间
    :param end_date: 公告查询截止时间
    :param search_key: 公告查询搜索关键字
    :returns: None 接口用于下载公告原文进行人工分析，不返回任何数据

    catetories::

        all         所有类别
        ndbg        年度报告
        bndbg       半年度报告
        yjdbg       一季度报告
        sjdbg       三季度报告
        scgkfx      首次公开发行及上市
        pg          配股
        zf          增发
        kzhz        可转换债券
        qzxg        权证相关公告
        qtrz        其他融资
        qyfpxzcs    权益及限制出售股份
        gqbd        股权变动
        jy          交易
        gddh        股东大会
        cqfxyj      澄清风险业绩预告
        tbclts      特别处理和退市
        bcgz        补充及更正
        zjjg        中介机构报告
        ssgszd      上市公司制度
        zqgg        债券公告
        qtzdsx      其它重大事项
        tzzgx       投资者关系信息
        dshgg       董事会公告
        jshgg       监事会公告

    """
    await _init(session)
    cate_list = [f"category_{cat}_szsh" for cat in categories]
    category_str = ";".join(cate_list) + ";"
    date_str = ""
    if start_date and end_date:
        date_str = f"{start_date}+~+{end_date}"

    page_size = 30
    page_cnt = 1
    page_num = 1

    logger.info("Downloading announcements data for %s", symbol)
    query_url = "http://www.cninfo.com.cn/information/companyinfo_n.html"
    await session.get(query_url)
    post_url = "http://www.cninfo.com.cn/cninfo-new/announcement/query"

    cninfo_symbol = symbol
    if cninfo_symbol.startswith("SZ") or cninfo_symbol.startswith("SH"):
        cninfo_symbol = cninfo_symbol[2:]
    post_form = {
        "category": category_str,
        "column": "",
        "limit": "",
        "pageNum": 1,
        "pageSize": page_size,
        "searchkey": search_key,
        "seDate": date_str,
        "sortName": "",
        "sortType": "",
        "stock": cninfo_symbol,
        "tabName": "fulltext",
    }

    announcements = []
    logger.info("Downloading announcements meta data for %s", symbol)
    while page_num <= page_cnt:
        post_form["pageNum"] = [page_num]
        try:
            async with session.post(post_url, data=post_form) as resp:
                if resp.status != 200:
                    logger.warning(
                        "Download announcement metadata from %s failed: http %s",
                        post_url,
                        resp.status,
                    )
                    break
                data = await resp.json()
                if "totalAnnouncement" in data:
                    total_cnt = data["totalAnnouncement"]
                elif "totalRecordNum" in data:
                    total_cnt = data["totalRecordNum"]
                else:
                    logger.warning(
                        "Download announcements meta data for %s failed: %s",
                        symbol,
                        data,
                    )
                    break
                page_cnt = total_cnt // page_num
                if total_cnt % page_num != 0:
                    page_cnt += 1
                if "announcements" in data:
                    announcements.extend(data["announcements"])
        except (aiohttp.ServerTimeoutError, asyncio.TimeoutError) as e:
            logger.warning(
                "Download announcements meta data for %s timeout: %s", symbol, e
            )
            continue
        page_num += 1

    if not announcements:
        logger.warning("no announcements data found for %s", symbol)
        return
    logger.info("Download announcements meta data for %s finished", symbol)

    symbol_data_dir = Path(data_path) / MODULE_DATA_DIR / symbol / "announcements"
    os.makedirs(symbol_data_dir, exist_ok=True)
    meta_file = symbol_data_dir / f"{symbol}_meta.json"
    with meta_file.open("w", encoding="utf-8") as dataf:
        json.dump(announcements, dataf, indent=4, ensure_ascii=False)

    aws = []
    logger.info("Downloading announcements files for %s", symbol)
    for announcement in announcements:
        logger.info(
            "Downloading %s from %s",
            announcement["announcementTitle"],
            announcement["adjunctUrl"],
        )
        annou_name = announcement.get("announcementTitle", None)
        if not annou_name:
            annou_name = announcement.get("announcementId", None)
        if not annou_name:
            annou_name = announcement.get("orgId", None)
        if not annou_name:
            logger.warning("No file name found, skipped: %s", announcement)
            continue
        annou_time = ""
        if "announcementTime" in announcement:
            annou_time = datetime.fromtimestamp(announcement["announcementTime"] / 1000)
            annou_time = annou_time.strftime("%Y%m%d%H%M%S")
        ftype = announcement.get("adjunctType", "raw")
        fpath = symbol_data_dir / f"{annou_name}-{annou_time}.{ftype}"

        url = "http://www.cninfo.com.cn/" + announcement["adjunctUrl"]
        aws.append(_get_one_announcement(session, url, fpath))
    await asyncio.gather(*aws, return_exceptions=True)
    logger.info("Download announcements files for %s finished", symbol)
    return None


@add_doc(async_get_announcements.__doc__)
def get_announcements(*args, **kwargs):
    ret = fetch_http_data(async_get_announcements, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option("-s", "--symbol", type=str, required=True)
@click.option(
    "-st",
    "--start",
    default=str(date.today().replace(month=1, day=1) - relativedelta(years=3)),
    show_default=True,
)
@click.option("-ed", "--end", default=str(date.today()), show_default=True)
@click.option(
    "-ct", "--category", default="ndbg,bndbg,yjdbg,sjdbg", show_default=True, type=str
)
@click.option("-se", "--search", default=True)
@click.option(
    "-f",
    "--file-path",
    "save_path",
    type=click.Path(exists=False)
)
@stock_cli_group.command("announce", short_help="获取公告原文")
@click.pass_context
def announcements_cli(ctx, symbol, start, end, save_path, category, search):
    """从cninfo获取公告原文

    symbol: 使用雪球网的代码格式

    announcements category

    \b
        all         所有类别
        ndbg        年度报告
        bndbg       半年度报告
        yjdbg       一季度报告
        sjdbg       三季度报告
        scgkfx      首次公开发行及上市
        pg          配股
        zf          增发
        kzhz        可转换债券
        qzxg        权证相关公告
        qtrz        其他融资
        qyfpxzcs    权益及限制出售股份
        gqbd        股权变动
        jy          交易
        gddh        股东大会
        cqfxyj      澄清风险业绩预告
        tbclts      特别处理和退市
        bcgz        补充及更正
        zjjg        中介机构报告
        ssgszd      上市公司制度
        zqgg        债券公告
        qtzdsx      其它重大事项
        tzzgx       投资者关系信息
        dshgg       董事会公告
        jshgg       监事会公告
    """
    start_dt = parse_dt(start, return_date=True)
    end_dt = parse_dt(end, return_date=True)

    if not save_path:
        save_path = ctx.obj["data_path"]
    data = get_announcements(
        symbol, category.split(","), save_path, start_dt, end_dt, search_key=""
    )
    if isinstance(data, Exception):
        raise data


if __name__ == "__main__":
    announcements_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""获取市场行情一览

信息获取通道包括：

    * https://xueqiu.com/stock/cata/stocklist.json
    * https://xueqiu.com/fund/quote/list.json

加载已保存的数据::

    import pandas as pd
    from pathlib import Path

    df = pd.read_json(Path('20181018165415.json')
    df.set_index(["symbol"], inplace=True)
    # df.drop_duplicates(inplace=True)
    df.apply(pd.to_numeric, errors="coerce") # convert to numertic types

    # df['day'] = date.today()
    # df["day"] = datetime.now().replace(hour=16, minute=0, second=0, microsecond=0)
    # set index
    # df.set_index(["symbol", "day"], inplace=True)
"""
import os
import time
import json
import copy
import asyncio
import logging
from pathlib import Path
from datetime import datetime

import click
import pandas as pd

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..config import get_config
from ..env import _init_in_session
from ..utils import fetch_http_data, add_doc


logger = logging.getLogger(__file__)
__all__ = ["async_get_list_qutes", "get_list_quotes"]


async def _init(session, force=False):
    if force or not _init_in_session.get("xueqiu"):
        _init_in_session["xueqiu"] = True
        await session.get("https://xueqiu.com")
    return True


async def async_get_list_qutes(
    session, data_type="stock", data_path=None, return_df=True
):
    """
    获取每日行情概览信息，只能获取当天的
    返回一个 pd.DataFrame
    出错，返回 None

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param data_type: stock: 沪深股票 cb: 可转债 eft: ETF基金 fenji: 分级基金
    :param data_path: 数据保存路径
    :param return_df: 是否返回 `pandas.DataFrame` 对象，False 返回原始数据

    :returns: 行情原始数据或带有行情数据的 `pandas.DataFrame` 对象，见 return_df 参数
    """
    await _init(session)

    quotes = []
    page_size = 90
    curr_page = page_cnt = 1
    params = {
        "_": 0,
        "order": "desc",
        "orderby": "percent",
        "page": curr_page,
        "size": page_size,
    }
    quotes_url = "https://xueqiu.com/stock/cata/stocklist.json"
    if data_type == "stock":
        params["type"] = "11,12"
    elif data_type == "cb":
        params["exchange"] = "CN"
        params["industry"] = "可转债"
    elif data_type == "etf":
        params["parent_type"] = 13
        params["type"] = 135
        params["orderBy"] = "percent"
        quotes_url = "https://xueqiu.com/fund/quote/list.json"
    elif data_type == "fenji":
        params["parent_type"] = 1
        params["type"] = 14
        params["orderBy"] = "percent"
        quotes_url = "https://xueqiu.com/fund/quote/list.json"

    logger.info("start download live quotes from xueqiu for %s...", data_type)

    async def fetch_one_page(url, params):
        async with session.get(url, params=params) as resp:
            if resp.status != 200:
                logger.warning(
                    "Download list_quotes from url %s failed: http %s", url, resp.status
                )
                return None
            return await resp.json()

    def process_one_page(resp_json, quotes):
        if not resp_json:
            return 0
        if data_type in ("stock", "cb"):
            if not resp_json["success"]:
                logger.error("Get daily quotes for %s failed: %s", data_type, resp_json)
            total_cnt = resp_json["count"]["count"]
        elif data_type in ("etf",):
            if "error_code" in resp_json:
                logger.error("Get daily quotes for %s failed: %s", data_type, resp_json)
            total_cnt = resp_json["count"]
        elif data_type in ("fenji",):
            if "error_code" in resp_json:
                logger.error("Get daily quotes for %s failed: %s", data_type, resp_json)
            total_cnt = resp_json["count"]
        quotes.extend(resp_json["stocks"])
        return total_cnt

    aws = []
    while curr_page <= page_cnt:
        params["page"] = curr_page
        params["_"] = int(time.time() * 1000)

        logger.info("Fetching first page")
        resp_json = await fetch_one_page(quotes_url, params)
        total_cnt = process_one_page(resp_json, quotes)

        page_cnt = total_cnt // page_size + 1 if total_cnt % page_size != 0 else 0
        logger.info("First page process finish, total_page: %s page_cnt")
        curr_page += 1

        if page_cnt > 2:
            for index in range(curr_page, page_cnt + 1):
                new_params = copy.copy(params)
                new_params["page"] = index
                aws.append(fetch_one_page(quotes_url, new_params))
            curr_page = index + 1

    if aws:
        pages = await asyncio.gather(*aws, return_exceptions=True)
        for page in pages:
            if not isinstance(page, Exception):
                process_one_page(page, quotes)

    if not quotes:
        logger.warn("no data downloaded for %s, return None", data_type)
    logger.info("download xueqiu daily quotes for %s finish", data_type)

    if data_path:
        data_path = Path(data_path) / MODULE_DATA_DIR / "list_quotes"
        os.makedirs(data_path, exist_ok=True)
        data_fname = data_type + "-" + datetime.now().strftime("%Y%m%d%H%M%S") + ".json"
        data_file = data_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(quotes, dataf, indent=4, ensure_ascii=False)

    if not return_df:
        return quotes

    df = pd.DataFrame(quotes)
    # df['day'] = date.today()
    # df["day"] = datetime.now().replace(hour=16, minute=0, second=0, microsecond=0)
    # set index
    # df.set_index(["symbol", "day"], inplace=True)
    df.set_index(["symbol"], inplace=True)
    df.drop_duplicates(inplace=True)
    # convert to numertic types
    df.apply(pd.to_numeric, errors="ignore")
    return df


@add_doc(async_get_list_qutes.__doc__)
def get_list_quotes(*args, **kwargs):
    ret = fetch_http_data(async_get_list_qutes, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option("-t", "--data-type", default="stock", show_default=True)
@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.option("-p/-np", "--print/--no-print", "show", default=True)
@stock_cli_group.command("list-quotes")
@click.pass_context
def list_quotes_cli(ctx, data_type, save_path, show):
    """从雪球获取最新的行情快照

    对于交易时间，可以获取实时的数据；
    非交易时间则获取上一交易日的数据。

    \b
    data_type
        stock: 沪深股票
        cb: 可转债
        eft: ETF基金
        fenji: 分级基金
    """
    if not save_path:
        save_path = ctx.obj["data_path"]
    data = get_list_quotes(data_type, save_path)
    if show:
        click.echo(data)


if __name__ == "__main__":
    list_quotes_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""数据模块

从网上抓取数据

同步提供数据抓取并返回接口及抓取并存储接口
"""
from . import mkt_calendar
from . import announcement
from . import fundamentals
from . import fundamentals_xq
from . import list_quotes
from . import hist_quotes
from . import live_quotes
from . import fenhong
from . import zengfa
from . import guben
from . import gudong
from . import insider
from . import picker_xq

from .mkt_calendar import *         # noqa
from .announcement import *         # noqa
from .fundamentals import *         # noqa
from .fundamentals_xq import *      # noqa
from .list_quotes import *          # noqa
from .hist_quotes import *          # noqa
from .live_quotes import *          # noqa
from .fenhong import *              # noqa
from .zengfa import *               # noqa
from .guben import *                # noqa
from .gudong import *               # noqa
from .insider import *              # noqa
from .picker_xq import *            # noqa


__all__ = (
    mkt_calendar.__all__
    + announcement.__all__
    + fundamentals.__all__
    + fundamentals_xq.__all__
    + fenhong.__all__
    + zengfa.__all__
    + guben.__all__
    + gudong.__all__
    + insider.__all__
    + picker_xq.__all__
    + list_quotes.__all__
    + live_quotes.__all__
    + hist_quotes.__all__
)



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""每日市场事件

获取每日市场事件，事件类型包括：

    * 停牌、复牌
    * 首发新股上市
    * 首发新股招股书刊登
    * 首发新股网上发行
    * 中签率公告
    * 授予股份上市
    * 股东资格登记
    * 股东大会召开
    * 网络投票开始、网络投票结束
    * 分红转增股权登记、分红转增除权除息、分红转增红利发放、分红转增股份上市
    * 增发新股招股书刊登

等，信息获取通道包括：

    http://www.cninfo.com.cn/cninfo-new/memo?queryDate=2017-07-19

加载已保存的数据::

    import json

    from pathlib import Path
    with Path("xxx.json").open(encoding="utf-8") as f:
        data = json.load(f)
"""
import os
import json
import asyncio
import logging
from pathlib import Path
from datetime import date, timedelta

import click
import aiohttp

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..utils import iter_dt, parse_dt, fetch_http_data, add_doc


__all__ = ["async_get_market_events", "get_market_events"]
logger = logging.getLogger(__file__)


async def _init(session, force=False):
    return True


async def async_get_market_events(session, start, end, data_path=None):
    """从巨潮信息网获取某天的交易日历

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param start: 日历数据开始时间
    :param end: 日历数据结束时间
    :param data_path: 数据保存路径

    :returns: 日历数据 `dict`
    """
    datas = {}
    await _init(session)

    async def fetch_one_day(day_str):
        url = "http://www.cninfo.com.cn/cninfo-new/memo/memoQuery"
        try:
            async with session.post(url, data={"queryDate": day_str}) as resp:
                datas[day_str] = await resp.json()
        except (aiohttp.ServerTimeoutError, asyncio.TimerHandle) as e:
            logger.warning("Download calendar event data for %s failed：%s.", day_str, e)

    aws = []
    for day in iter_dt(start, end, include_end=True):
        day_str = day.strftime("%Y-%m-%d")
        aws.append(fetch_one_day(day_str))
    await asyncio.gather(*aws, return_exceptions=True)
    logger.info("calendar event data from %s to %s has been downloaded.", start, end)

    if data_path:
        file_path = Path(data_path) / MODULE_DATA_DIR / "market_events"
        os.makedirs(file_path, exist_ok=True)
        data_file = (
            file_path / f"{start.strftime('%Y%m%d')}-{end.strftime('%Y%m%d')}.json"
        )
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(datas, dataf, indent=4, ensure_ascii=False)
        logger.info("calendar data has been saved to: %s", data_file)
    return datas


@add_doc(async_get_market_events.__doc__)
def get_market_events(*args, **kwargs):
    ret = fetch_http_data(async_get_market_events, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option(
    "-st", "--start", default=str(date.today() - timedelta(days=30)), show_default=True
)
@click.option("-ed", "--end", default=str(date.today()), show_default=True)
@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.option("-p/-np", "--print/--no-print", "show", default=True)
@stock_cli_group.command("cale")
@click.pass_context
def market_events_cli(ctx, start, end, save_path, show):
    """从cninfo获取日历事件"""
    start_dt = parse_dt(start)
    end_dt = parse_dt(end)
    if not save_path:
        save_path = ctx.obj["data_path"]
    data = get_market_events(start_dt, end_dt, save_path)
    if show:
        click.echo(json.dumps(data, indent=2, ensure_ascii=False))


if __name__ == "__main__":
    market_events_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from pathlib import Path


_config = None
def_config_path = Path("~/.config/fintie")
logger = logging.getLogger(__name__)


def load_config(config_file=def_config_path):
    global _config
    config_file = Path(config_file)
    if not config_file.exists():
        logger.warning("配置文件(%s)不存在，配置设置为空", config_file)
        _config = {}
        return

    with config_file.open(encoding="utf-8") as f:
        _config = json.load(f)


def get_config(key, default=None):
    if _config is None:
        load_config()
    return _config.get(key, default)


def set_config(key, val):
    if _config is None:
        load_config()
    _config[key] = val
    return True



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import asyncio

from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as parse_datetime

from fintie.env import get_http_session


logger = logging.getLogger(__name__)


def convert_number(num_str, cls=int):
    """带逗号格式的数值字符串转为数值类型

    :params num_str: 数值字符串
    :params cls: 数值类型，默认为 `int`
    :returns: cls 类型的数值
    """
    return cls(num_str.replace(",", ""))


def iter_dt(start_dt, end_dt, include_start=True, include_end=False, **kwargs):
    """按步进值迭代一个日期范围的 Generator ，默认步进值是一天

    **NOTICE** 如果需要迭代的步进值小于一天，请传递 `datetime` 类型

    :params start_date: 开始日期
    :params end_date: 结束日期
    :params include_start: 是否包含开始时间
    :params include_end: 是否包含结束时间
    :params kwargs: 传递给 `dateutil.relativedelta` 的参数，用于计算步进值
    """
    if not kwargs:
        kwargs = {"days": 1}

    current_dt = start_dt
    if not include_start:
        current_dt += relativedelta(**kwargs)
    if include_end:
        end_dt += relativedelta(**kwargs)

    while current_dt < end_dt:
        yield current_dt
        current_dt += relativedelta(**kwargs)


def parse_dt(date_str, return_date=False):
    """解析字符串为 Date 类型，失败报 TypeError/ValueError 异常

    :params date_str: 要解析的时间日期字符串
    :params return_date: 如果为 `True` 返回 `datetime.date` 类型，否则返回 `datetime.datetime`
    :returns: `datetime` 如果 return_date 是 `False` ，否则返回 `date`
    """
    dtime = parse_datetime(date_str)
    if return_date:
        return dtime.date()
    return dtime


async def wrap_session_run(func, *args, **kwargs):
    async with get_http_session(force=True) as session:
        ret = await func(session, *args, **kwargs)
        return ret


def async2sync_run(*aws, return_exceptions=True):
    """将异步函数转为同步函数

    :params aws: 异步协程列表
    :params return_exceptions: 是否运行返回异常，不 允许的话任何一个协程异常都会导致本函数异常
    :returns: 一个列表包含了所有协程的返回值
    """
    loop = asyncio.get_event_loop()
    if loop.is_running():
        logger.warning("EventLoop 已经在运行，请调用异步接口！（或许您是在新版的 Notebook 中？）")
        return None
    # TODO: use ayncio.run for Python 3.7
    return loop.run_until_complete(
        asyncio.gather(*aws, return_exceptions=return_exceptions)
    )


def fetch_http_data(func, *args, **kwargs):
    """将异步的http取数据接口转为同步方式"""
    return async2sync_run(wrap_session_run(func, *args, **kwargs))[0]


def add_doc(doc):
    """一个给函数添加文档字符串的装饰器函数"""
    def func_wrapper(func):
        func.__doc__ = doc
        return func
    return func_wrapper



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""数据模块命令行"""
import os
import logging

import click

from .config import load_config, set_config, get_config, def_config_path


class Env(object):
    """Global context object"""
    pass


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option(
    "-c",
    "--conf",
    type=click.Path(),
    help="配置文件",
    default=def_config_path,
    show_default=True,
)
@click.option("-d", "--data-path", type=click.Path(), help="数据存储路径，会覆盖 conf 文件里的设置")
@click.pass_context
def cli(ctx, verbose, conf, data_path):
    """Console script for fintie

    Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Project url: https://github.com/qytz/fintie
    """
    log_levels = [
        logging.CRITICAL,
        logging.ERROR,
        logging.WARNING,
        logging.INFO,
        logging.DEBUG,
    ]
    if verbose >= len(log_levels):
        verbose = len(log_levels) - 1
    log_fmt = (
        "%(asctime)-15s %(levelname)s %(filename)s %(lineno)d %(process)d %(message)s"
    )
    logging.basicConfig(
        filename="fintie.log", level=log_levels[verbose], format=log_fmt
    )

    if conf:
        load_config(conf)

    if data_path:
        set_config("data_path", data_path)
    else:
        data_path = get_config("data_path", os.path.join(os.getcwd(), "fintie_data"))
    ctx.ensure_object(dict)
    ctx.obj["data_path"] = data_path
    click.echo(f"下载的数据将会保存到 {data_path}")


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""提供股本信息查询

信息获取通道包括：

    * 当日财务指标    http://xueqiu.com/S/SH600559/MRCWZB
    * 股票收益率指标  http://xueqiu.com/S/SH600559/GPSYLZB

    * 主要财务指标    https://xueqiu.com/S/SH600559/ZYCWZB
    * 单季财务指标    http://xueqiu.com/S/SH600559/DJCWZB
    * 综合损益表      http://xueqiu.com/S/SH600559/GSLRB
    * 资产负债表      http://xueqiu.com/S/SH600559/ZCFZB
    * 现金流量表      http://xueqiu.com/S/SH600559/XJLLB


加载已保存的数据::

    # 主要财务指标/单季财务指标/综合损益表/资产负债表/现金流量表

    import pandas as pd
    from pathlib import Path

    df = pd.read_json(Path('xxx.json')
    # optional, set index
    if "reportdate" if df:
        df.set_index("reportdate", inplace=True)
    else:
        df.set_index("enddate", inplace=True)
        or
        df.set_index("begindate", inplace=True)

    # 当日财务指标/股票收益率指标
    import json

    from pathlib import Path
    with Path("xxx.json").open(encoding="utf-8") as f:
        data = json.load(f)
"""
import os
import time
import json
import asyncio
import logging
from pathlib import Path
from datetime import date

import click
import pandas as pd

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..env import _init_in_session
from ..utils import fetch_http_data, add_doc


logger = logging.getLogger(__file__)
__all__ = ["async_get_funda", "get_funda"]
FUNDA_TABLES = {
    # 当日财务指标
    "MRCWZB": "https://xueqiu.com/stock/f10/dailypriceextend.json",
    # 股票收益率指标
    "GPSYLZB": "https://xueqiu.com/stock/f10/yieldindic.json",
    # 主要财务指标
    "ZYCWZB": "https://xueqiu.com/stock/f10/finmainindex.json",
    # 单季财务指标
    "DJCWZB": "https://xueqiu.com/stock/f10/finqindic.json",
    # 综合损益表
    "GSLRB": "https://xueqiu.com/stock/f10/incstatement.json",
    # 资产负债表
    "ZCFZB": "https://xueqiu.com/stock/f10/balsheet.json",
    # 现金流量表
    "XJLLB": "https://xueqiu.com/stock/f10/cfstatement.json",
}


async def _init(session, force=False):
    if force or not _init_in_session.get("xueqiu"):
        _init_in_session["xueqiu"] = True
        await session.get("https://xueqiu.com")
    return True


async def async_get_funda(session, symbol, table, data_path=None, return_df=True):
    """
    从雪球获取财务数据

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param table: 财务数据类型

                  MRCWZB    # 当日财务指标
                  GPSYLZB   # 股票收益率指标
                  ZYCWZB    # 主要财务指标
                  DJCWZB    # 单季财务指标
                  GSLRB     # 综合损益表
                  ZCFZB     # 资产负债表
                  XJLLB     # 现金流量表

    :param data_path: 数据保存路径
    :param return_df: 是否返回 `pandas.DataFrame` 对象，False 返回原始数据

    :returns: 原始数据或 `pandas.DataFrame` 对象，见 return_df 参数，
              如果是 MRCWZB/GPSYLZB table，直接返回字典，reture_df 参数无效
              失败则返回 `None`
    """
    assert table in FUNDA_TABLES
    await _init(session)

    page_size = 10000
    curr_page = 1
    params = {"_": 0, "symbol": symbol, "page": curr_page, "size": page_size}

    logger.info("start download funda2 from xueqiu for %s...", table)
    url = FUNDA_TABLES[table]
    params["_"] = int(time.time() * 1000)

    list_data = False
    date_str = str(date.today())
    async with session.get(url, params=params) as resp:
        if resp.status != 200:
            logger.warning("get funda2 from %s failed: %s", url, resp.status)
            return None
        funda_data = await resp.json()
        if "list" in funda_data:
            list_data = True
            funda_data = funda_data.get("list")

    if not funda_data:
        logger.warn("no funda2 data downloaded for %s from %s, return None", table, url)
        return None
    logger.info("download funda2 table %s from %s finish", table, url)
    if data_path:
        data_path = Path(data_path) / MODULE_DATA_DIR / "funda2"
        os.makedirs(data_path, exist_ok=True)
        data_fname = "-".join((symbol, table, date_str)) + ".json"
        data_file = data_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(funda_data, dataf, indent=4, ensure_ascii=False)

    if not return_df or not list_data:
        return funda_data

    df = pd.DataFrame(funda_data)
    # set index
    # try:
    #     df.set_index("reportdate", inplace=True)
    # except KeyError:
    #     df.set_index("enddate", inplace=True)
    return df


@add_doc(async_get_funda.__doc__)
def get_funda(*args, **kwargs):
    ret = fetch_http_data(async_get_funda, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option("-s", "--symbol", required=True)
@click.option(
    "-t",
    "--table",
    type=click.Choice(FUNDA_TABLES),
    default="MRCWZB",
    show_default=True,
)
@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.option("-p/-np", "--print/--no-print", "show", default=True)
@stock_cli_group.command("funda2")
@click.pass_context
def funda2_cli(ctx, symbol, table, save_path, show):
    """从雪球获取财报数据"""
    if not save_path:
        save_path = ctx.obj["data_path"]
    data = get_funda(symbol, table, save_path)
    if show:
        click.echo(data)


if __name__ == "__main__":
    funda2_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""本模块负责获取股票的历史交易行情信息。

获取通道包括：

    * 基础信息，包含最新行情快照

      https://stock.xueqiu.com/v5/stock/quote.json?symbol=SZ002353&extend=detail

    * 最近的 100 条成交记录

      https://stock.xueqiu.com/v5/stock/history/trade.json?symbol=SZ002353&count=100

    * 实时-盘口

      https://stock.xueqiu.com/v5/stock/realtime/pankou.json?symbol=SZ002353


数据加载

    trades::

        import json
        from pathlib import Path

        import pandas as pd

        with Path('SZ002353-trade-20181019201938.json').open(encoding="utf-8") as f:
            quotes = json.load(f)

        df = pd.DataFrame(data=quotes["items"])
        df.timestamp = pd.to_datetime(df.timestamp, unit="ms")
        df.set_index("timestamp", inplace=True)


TODO:

    * add default http headers

      headers = {"X-Forwarded-For": ipAddress}
"""
import os
import json
import logging
from pathlib import Path
from datetime import datetime

import click
import pandas as pd

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..env import _init_in_session
from ..utils import fetch_http_data, add_doc


__all__ = [
    "async_get_trade_info",
    "async_get_pankou",
    "async_get_live_info",
    "get_trade_info",
    "get_pankou",
    "get_live_info",
]
logger = logging.getLogger(__file__)
INIT_URLS = ["https://xueqiu.com"]
QUOTE_TYPES = ("pankou", "trades", "live-info")


async def _init(session, force=False):
    if force or not _init_in_session.get("xueqiu"):
        _init_in_session["xueqiu"] = True
        await session.get("https://xueqiu.com")
    return True


async def async_get_trade_info(session, symbol, data_path=None, return_df=True):
    """获取最近的交易记录

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param data_path: 数据保存路径
    :param return_df: 是否返回 `pandas.DataFrame` 对象，False 返回原始数据

    :returns: 行情原始数据或带有行情数据的 `pandas.DataFrame` 对象，见 return_df 参数
    """
    await _init(session)

    url = "https://stock.xueqiu.com/v5/stock/history/trade.json"
    params = {"symbol": symbol}
    async with session.get(url, params=params) as resp:
        if resp.status != 200:
            logger.warning("get live trades from %s failed: %s", url, resp.status)
            return None
        data_json = await resp.json()
        if data_json.get("error_code", -1) != 0:
            logger.warning(
                "get live trades from %s failed, error_code: %s", url, resp.status
            )
            return None
        quotes = data_json["data"]

    if data_path:
        data_path = Path(data_path) / MODULE_DATA_DIR / symbol / "live_quotes"
        os.makedirs(data_path, exist_ok=True)
        try:
            ref_dt = datetime.fromtimestamp(quotes["items"][0]["timestamp"] / 1000)
        except (KeyError, TypeError):
            ref_dt = datetime.now()

        data_fname = (
            "-".join((symbol, "trade", ref_dt.strftime("%Y%m%d%H%M%S"))) + ".json"
        )
        data_file = data_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(quotes, dataf, indent=4, ensure_ascii=False)

    if not return_df:
        return quotes

    df = pd.DataFrame(data=quotes["items"])
    df.timestamp = pd.to_datetime(df.timestamp, unit="ms")
    df.set_index("timestamp", inplace=True)
    return df


async def async_get_pankou(session, symbol, data_path=None):
    """获取最新的盘口数据
    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param data_path: 数据保存路径

    :returns: 盘口数据 `dict`
    """
    await _init(session)

    url = "https://stock.xueqiu.com/v5/stock/realtime/pankou.json"
    params = {"symbol": symbol}
    async with session.get(url, params=params) as resp:
        if resp.status != 200:
            logger.warning("get live pankou from %s failed: %s", url, resp.status)
            return None
        data_json = await resp.json()
        if data_json.get("error_code", -1) != 0:
            logger.warning(
                "get live pankou from %s failed, error_code: %s", url, resp.status
            )
            return None
        quotes = data_json["data"]

    if data_path:
        data_path = Path(data_path) / MODULE_DATA_DIR / symbol / "live_quotes"
        os.makedirs(data_path, exist_ok=True)
        try:
            ref_dt = datetime.fromtimestamp(quotes["timestamp"] / 1000)
        except (KeyError, TypeError):
            ref_dt = datetime.now()

        data_fname = (
            "-".join((symbol, "pankou", ref_dt.strftime("%Y%m%d%H%M%S"))) + ".json"
        )
        data_file = data_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(quotes, dataf, indent=4, ensure_ascii=False)

    return quotes


async def async_get_live_info(session, symbol, data_path=None):
    """获取最新的基本信息，包括最新价格市值等

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param data_path: 数据保存路径

    :returns: 基本信息数据 `dict`
    """
    await _init(session)

    url = "https://stock.xueqiu.com/v5/stock/quote.json"
    params = {"symbol": symbol, "extend": "detail"}
    async with session.get(url, params=params) as resp:
        if resp.status != 200:
            logger.warning("get live trade info from %s failed: %s", url, resp.status)
            return None
        data_json = await resp.json()
        if data_json.get("error_code", -1) != 0:
            logger.warning(
                "get live trade info from %s failed, error_code: %s", url, resp.status
            )
            return None
        quotes = data_json["data"]

    if data_path:
        data_path = Path(data_path) / MODULE_DATA_DIR / symbol / "live_quotes"
        os.makedirs(data_path, exist_ok=True)
        try:
            ref_dt = datetime.fromtimestamp(quotes["quote"]["timestamp"] / 1000)
        except (KeyError, TypeError):
            ref_dt = datetime.now()

        data_fname = (
            "-".join((symbol, "quotes", ref_dt.strftime("%Y%m%d%H%M%S"))) + ".json"
        )
        data_file = data_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(quotes, dataf, indent=4, ensure_ascii=False)

    return quotes


@add_doc(async_get_trade_info.__doc__)
def get_trade_info(*args, **kwargs):
    ret = fetch_http_data(async_get_trade_info, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@add_doc(async_get_pankou.__doc__)
def get_pankou(*args, **kwargs):
    ret = fetch_http_data(async_get_pankou, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@add_doc(async_get_live_info.__doc__)
def get_live_info(*args, **kwargs):
    ret = fetch_http_data(async_get_live_info, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option("-s", "--symbol", required=True)
@click.option(
    "-t", "--type", "quotes_type", type=click.Choice(QUOTE_TYPES), default="pankou"
)
@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.option("-p/-np", "--print/--no-print", "show", default=True)
@stock_cli_group.command("live-quotes")
@click.pass_context
def live_quotes_cli(ctx, symbol, quotes_type, save_path, show):
    """从雪球获取实时行情"""
    if quotes_type == "trades":
        data = get_trade_info(symbol, save_path)
    elif quotes_type == "info":
        data = get_live_info(symbol, save_path)
    elif quotes_type == "pankou":
        data = get_pankou(symbol, save_path)
    else:
        click.echo("quote_type invalid: %s", quotes_type)
        return -1

    if not save_path:
        save_path = ctx.obj["data_path"]

    if show:
        if isinstance(data, (list, dict)):
            click.echo(json.dumps(data, indent=4, ensure_ascii=False))
        else:
            click.echo(data)


if __name__ == "__main__":
    live_quotes_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" 本模块负责获取上市公司财务数据，包含利润、资产负债、现金流量三大表及摘要和主要财务指标。

信息获取通道包括：

    * http://quotes.money.163.com/service/zcfzb_000333.html
    * http://quotes.money.163.com/service/lrb_000333.html
    * http://quotes.money.163.com/service/xjllb_000333.html
    * http://quotes.money.163.com/service/cwbbzy_000333.html
    * http://quotes.money.163.com/service/zycwzb_000333.html?type=report

加载已保存数据::

    import pandas as pd
    from pathlib import Path
    df = pd.read_csv(Path("xxx.json"), index_col=0)
"""
import os
import logging
import asyncio
from pathlib import Path

import click
import aiohttp
import pandas as pd

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..utils import fetch_http_data, add_doc


logger = logging.getLogger(__file__)
__all__ = [
    "get_funda_tab",
    "async_get_funda_tab",
    "get_fundamentals",
    "async_get_fundamentals",
]


async def _init(session):
    return True


async def async_get_funda_tab(session, symbol, tab_name, return_df=True):
    """获取某表的财务数据

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param tab_name: lrb/zcfzb/xjllb/cwbbzy/zycwzb
    :param data_path: 数据保存路径
    :param return_df: 是否返回 `pandas.DataFrame` 对象，False 返回原始数据

    :returns: 原始财务数数据或带有财务数据的 `pandas.DataFrame` 对象，见 return_df 参数
    """
    assert tab_name in ("lrb", "zcfzb", "xjllb", "cwbbzy", "zycwzb")

    await _init(session)
    symbol_163 = symbol
    if symbol_163.startswith("SZ") or symbol_163.startswith("SH"):
        symbol_163 = symbol[2:]
    url = f"http://quotes.money.163.com/service/{tab_name}_{symbol_163}.html"
    try:
        async with await session.get(url) as resp:
            if resp.status != 200:
                logger.warning(
                    "Download funda for %s.%s from url failed: http %s",
                    symbol,
                    tab_name,
                    url,
                    resp.status,
                )
                return None
            data = await resp.text()
            encoding = resp.get_encoding()
    except (aiohttp.ServerTimeoutError, asyncio.TimerHandle):
        logger.warning("get %s for %s failed: timeout", tab_name, symbol)

    if not return_df:
        return data
    return pd.read_csv(data, encoding=encoding, index_col=0)


async def async_get_fundamentals(session, symbols, data_path):
    """获取财务数据

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param data_path: 数据保存路径

    :returns: Nothing
    """

    async def get_one_funda(symbol, tab_name, path):
        data = await async_get_funda_tab(session, symbol, tab_name, return_df=False)
        save_file = path / (tab_name + ".csv")
        with save_file.open("w", encoding="utf-8") as f:
            f.write(data)

    await _init(session)
    aws = []
    logger.info("Getting fundamental from 163")
    for symbol in symbols:
        symbol_data_dir = Path(data_path) / MODULE_DATA_DIR / symbol / "fundamental"
        os.makedirs(symbol_data_dir, exist_ok=True)
        for tab_name in ("lrb", "zcfzb", "xjllb", "cwbbzy", "zycwzb"):
            aws.append(get_one_funda(symbol, tab_name, symbol_data_dir))
    rets = await asyncio.gather(*aws, return_exceptions=True)
    for ret in rets:
        if isinstance(ret, Exception):
            raise ret

    return True


@add_doc(async_get_funda_tab.__doc__)
def get_funda_tab(*args, **kwargs):
    ret = fetch_http_data(async_get_funda_tab, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@add_doc(async_get_fundamentals.__doc__)
def get_fundamentals(*args, **kwargs):
    ret = fetch_http_data(async_get_fundamentals, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.argument("symbols", nargs=-1, required=True)
@stock_cli_group.command("funda")
@click.pass_context
def fundamental_cli(ctx, symbols, save_path):
    """从163获取财务报表及财务指标信息"""
    if not save_path:
        save_path = ctx.obj["data_path"]
    get_fundamentals(symbols, save_path)


if __name__ == "__main__":
    fundamental_cli()



# -*- coding: utf-8 -*-
# This file is part of fintie.

# Copyright (C) 2018-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""行情获取模块

本模块负责获取股票的历史交易行情信息。

获取通道包括：

    Get https://stock.xueqiu.com/v5/stock/chart/kline.json

    params::

        "symbol": 代码
        "begin": 时间戳
        "period": 频率,
        "type": 复权类型,
        "count": 数量,
        "indicator": "kline,ma,macd,kdj,boll,rsi,wr,bias,cci,psy",

加载::

    from pathlib import Path
    import pandas as pd

    with Path('data.json').open(encoding="utf-8") as f:
        quotes = json.load(f)

    df = pd.DataFrame(data=quotes['item'], column=quotes['column'])
    df.timestamp = pd.to_datetime(df.timestamp, unit='ms')
    df.set_index('timestamp', inplace=True)

TODO

    * add default http headers

      headers = {"X-Forwarded-For": ipAddress}

    * 163 - 日K线数据

      Get http://quotes.money.163.com/service/chddata.html?code=1000333&start=20130918&end=20180803&fields=TCLOSE;HIGH;LOW;TOPEN;LCLOSE;CHG;PCHG;TURNOVER;VOTURNOVER;VATURNOVER;TCAP;MCAP
"""
import os
import json
import logging
from pathlib import Path
from datetime import datetime

import click
import pandas as pd

from .cli import stock_cli_group, MODULE_DATA_DIR
from ..env import _init_in_session
from ..utils import parse_dt, fetch_http_data, add_doc


logger = logging.getLogger(__file__)
__all__ = ["async_get_hist_quotes", "get_hist_quotes"]


SUPPORTED_FREQ = (
    "1m",
    "5m",
    "15m",
    "30m",
    "60m",
    "120m",
    "day",
    "week",
    "month",
    "quarter",
    "year",
)
FQ_TYPE = ("before", "after", "normal")


async def _init(session, force=False):
    if force or not _init_in_session.get("xueqiu"):
        _init_in_session["xueqiu"] = True
        await session.get("https://xueqiu.com")
    return True


async def async_get_hist_quotes(
    session,
    symbol,
    ref_dt,
    count=100,
    freq="1m",
    fq_type="before",
    data_path=None,
    return_df=True,
):
    """
    小于 day 频率的数据，会有时间限制，只能取最近的数据，请谨慎使用

    :param session: `aiohttp.ClientSession` 对象，同步接口不需要传
    :param symbol: 股票代码
    :param ref_dt: 行情数据参考日期，count 传负是截止日期，传正为开始日期
    :param count: 要取的行情数据的条数
    :param freq: 数据频率：1m/5m/15m/30m/60m/120m/day/week/month/quarter/year
    :param fq_type: before/after/normal 前复权、后复权、不复权
    :param data_path: 数据保存路径
    :param return_df: 是否返回 `pandas.DataFrame` 对象，False 返回原始数据

    :returns: 行情原始数据或带有行情数据的 `pandas.DataFrame` 对象，见 return_df 参数
    """
    assert freq in SUPPORTED_FREQ
    await _init(session)
    url = "https://stock.xueqiu.com/v5/stock/chart/kline.json"
    params = {
        "symbol": "SZ002353",
        "begin": int(ref_dt.timestamp()) * 1000,
        "period": freq,
        "type": fq_type,
        "count": count,
        "indicator": "kline,ma,macd,kdj,boll,rsi,wr,bias,cci,psy",
    }
    async with session.get(url, params=params) as resp:
        if resp.status != 200:
            logger.warning("get history quotes from %s failed: %s", url, resp.status)
            return None
        data = await resp.json()
        if data.get("error_code", -1) != 0:
            logger.warning("get history quotes from %s failed: %s", url, resp.status)
            return None
        quotes = data.get("data", {})

    if data_path:
        file_path = Path(data_path) / MODULE_DATA_DIR / symbol / "hist_quotes"
        os.makedirs(file_path, exist_ok=True)
        data_fname = (
            "-".join(
                (symbol, freq, ref_dt.strftime("%Y%m%d%H%M%S"), str(count), fq_type)
            )
            + ".json"
        )
        data_file = file_path / data_fname
        with data_file.open("w", encoding="utf-8") as dataf:
            json.dump(quotes, dataf, indent=4, ensure_ascii=False)

    if not return_df:
        return quotes

    df = pd.DataFrame(data=quotes["item"], columns=quotes["column"])
    df.timestamp = pd.to_datetime(df.timestamp, unit="ms")
    df.set_index("timestamp", inplace=True)
    return df


@add_doc(async_get_hist_quotes.__doc__)
def get_hist_quotes(*args, **kwargs):
    ret = fetch_http_data(async_get_hist_quotes, *args, **kwargs)
    if isinstance(ret, Exception):
        raise ret
    return ret


@click.option("-s", "--symbol", required=True)
@click.option(
    "-ed", "--end-dt", default=str(datetime.now()), show_default=True, help="行情截止时间"
)
@click.option("-c", "--count", default=200, show_default=True, help="行情条数")
@click.option(
    "-fq",
    "--freq",
    default="day",
    type=click.Choice(SUPPORTED_FREQ),
    show_default=True,
    help="行情的频率",
)
@click.option(
    "-ft",
    "--fq-type",
    default="before",
    type=click.Choice(FQ_TYPE),
    show_default=True,
    help="复权类型",
)
@click.option(
    "-f",
    "--save-path",
    type=click.Path(exists=False)
)
@click.option("-p/-np", "--print/--no-print", "show", default=True)
@stock_cli_group.command("hist-quotes")
@click.pass_context
def hist_quotes_cli(ctx, symbol, end_dt, count, freq, fq_type, save_path, show):
    """从雪球获取历史行情数据

    day 以下的 freq 会有条数和时间限制，只能获取最近的数据
    """
    if not save_path:
        save_path = ctx.obj["data_path"]
    end_dt = parse_dt(end_dt)
    data = get_hist_quotes(symbol, end_dt, -count, freq, fq_type, save_path)
    if show:
        click.echo(data)


if __name__ == "__main__":
    hist_quotes_cli()
